import json
import os

from Empiric.internal.StatisticsTools import StatisticsTools

class GeneralSettings():
  _pathCollectedData = 'collected-data/'
  _generalSettings = 'settings.json'
  def __init__(self):
    if not os.path.exists(GeneralSettings._pathCollectedData):
      os.makedirs(GeneralSettings._pathCollectedData)
    if os.path.exists(self._filepath()):
      with open(self._filepath(), 'r') as f:
        self._generalSettings = json.load(f)
    else:
      self._generalSettings = {
        'statistics': {},
      }
  def _filepath(self):
    if not os.path.exists(self._pathCollectedData):
      os.makedirs(self._pathCollectedData)
    return os.path.join(GeneralSettings._pathCollectedData, GeneralSettings._generalSettings)
  def get(self, key=None):
    if key is None:
      return self._generalSettings
    else:
      return self._generalSettings[key] if key in self._generalSettings else None
  def setStatistics(self, statistics):
    statistics = StatisticsTools.fixKeyTitle(statistics)
    if 'key' in statistics:
      if statistics['key'] not in self._generalSettings['statistics']:
        self._generalSettings['statistics'][statistics['key']] = statistics
      return statistics['key']
    return None
  def save(self):
    with open(self._filepath(), 'w') as f:
      json.dump(self._generalSettings, f)
    return self._generalSettings



class AGGREGATE:
  COUNT = lambda xs: [len(x) for x in xs]
  FIRST = lambda xs: [x[0] for x in xs]
  MAX = lambda xs: [max(x) for x in xs]
  MEAN = lambda xs: [sum(x) / len(x) for x in xs]
  MIN = lambda xs: [min(x) for x in xs]
  SUM = lambda xs: [sum(x) for x in xs]
  COLLECT = lambda xss: [x for xs in xss for x in xs]
  @staticmethod
  def _fromString(s, xs):
    if s.upper() in xs.keys():
      return xs[s.upper()]
    return None
  @staticmethod
  def fromString(s):
    return AGGREGATE._fromString(s, {
      'COUNT': AGGREGATE.COUNT,
      'FIRST': AGGREGATE.FIRST,
      'MAX': AGGREGATE.MAX,
      'MEAN': AGGREGATE.MEAN,
      'MIN': AGGREGATE.MIN,
      'SUM': AGGREGATE.SUM,
      'COLLECT': AGGREGATE.COLLECT,
    })
  @staticmethod
  def needsFilterFromString(s):
    return AGGREGATE._fromString(s, {
      'COUNT': False,
      'FIRST': True,
      'MAX': True,
      'MEAN': True,
      'MIN': True,
      'SUM': True,
      'COLLECT': True,
    })

class VISUALIZATION_TYPE:
  BAR_CHART = 'BAR_CHART'
  BOX_PLOT = 'BOX_PLOT'
  TEXT_COLLECTION = 'TEXT_COLLECTION'
  MAP = 'MAP'



import copy
import json
from jsonpath_ng import parse
import os

from Empiric.internal.GeneralSettings import GeneralSettings
from Empiric.internal.ManuscriptMemory import ManuscriptMemory
from Empiric.Statistics import AGGREGATE, VISUALIZATION_TYPE
from Empiric.internal.StatisticsTools import StatisticsTools

class Tools:
  @staticmethod
  def _flatten(xss):
    return [x for xs in xss for x in xs]
  @staticmethod
  def apply(x, selector):
    for y in parse(selector).find(x):
      return y.value
    return None
  @staticmethod
  def find(x, selector, fFilter):
    return [match.value for match in parse(selector).find(x) if fFilter(match.value)]
  @staticmethod
  def filterFn(xs, f):
    return list(filter(f, xs))
  @staticmethod
  def map(xs, selector):
    return Tools.mapFn(xs, lambda x: [y.value for y in parse(selector).find(x)])
  @staticmethod
  def mapFn(xs, f):
    return [f(x) for x in xs]

class Statistics:
  @staticmethod
  def _flattenStatistics(statistics):
    def transform(s):
      if 'substatistics' not in s:
        return [s]
      result = []
      s2 = copy.deepcopy(s)
      del s2['substatistics']
      for sub in s['substatistics']:
        sub = StatisticsTools.fixKeyTitle(sub)
        if 'title' in sub:
          sub['subtitle'] = sub['title']
          del sub['title']
        if 'key' in sub:
          sub['subkey'] = sub['key']
          del sub['key']
        result.append({
          **s2,
          **sub,
        })
      return result
    result = {}
    for s in Tools._flatten([transform(s) for s in statistics.values()]):
      if s['key'] not in result:
       result[s['key']] = []
      result[s['key']].append(s)
    return result
  @staticmethod
  def _readData():
    data = []
    for filename in sorted(os.listdir(ManuscriptMemory._pathCollectedData)):
      if filename.endswith('.json'):
        with open(os.path.join(ManuscriptMemory._pathCollectedData, filename), 'r') as f:
          data.append(json.load(f))
    return data
  @staticmethod
  def _computeStatisticsKeys(data):
    statisticKeys = []
    statisticKeysAll = [list(enumerate([match.value for match in parse('$.memory.*.statistics').find(d)])) for d in data]
    statisticKeysAll = [[x for x in ska if x[1] is not None] for ska in statisticKeysAll]
    m = max(*[len(ska) for ska in statisticKeysAll]) + 1
    tmp = [[] for _ in range(m + 1)]
    for ska in statisticKeysAll:
      for i, s in ska:
        tmp[i].append(s)
    while len(tmp) > 0:
      tmp = [ska for ska in tmp if len(ska) > 0]
      if len(tmp) > 0:
        s = tmp[0][0]
        statisticKeys.append(s)
        tmp = [[x for x in ska if x != s] for ska in tmp]
    return statisticKeys
  @staticmethod
  def _prepareData(data, statisticsKeys, statistics):
    result = []
    for s in statisticsKeys:
      if s not in statistics:
        continue
      for stat in statistics[s]:
        if 'data' not in stat:
          continue
        sd = stat['data']
        if 'selector' not in sd or not ('aggregateByPage' in sd or ('aggregateByKey' in sd and 'value' in sd)):
          continue
        xs = []
        for d in data:
          x = Tools.find(d, '$.memory.*', lambda v: v['statistics'] == s)
          x = Tools.mapFn(x, lambda v: v['result'])
          xs.extend(x)
        if 'aggregateByPage' in sd:
          def _collect(xs, selector, aggregateByPage, defaultValue):
            xs = Tools.map(xs, '$.' + selector)
            if defaultValue is not None:
              xs = [(x if len(x) > 0 else [defaultValue]) for x in xs]
            elif AGGREGATE.needsFilterFromString(aggregateByPage):
              xs = Tools.filterFn(xs, lambda x: len(x) > 0)
            return AGGREGATE.fromString(aggregateByPage)(xs)
          if isinstance(sd['selector'], str):
            value = _collect(xs, sd['selector'], sd['aggregateByPage'], sd['defaultValue'] if 'defaultValue' in sd else None)
          else:
            value = {}
            for k, selector in sd['selector'].items():
              value[k] = _collect(xs, selector, sd['aggregateByPage'], sd['defaultValue'][k] if 'defaultValue' in sd and k in sd['defaultValue'] else None)
          result.append({
            'key': stat['key'] if 'key' in stat else None,
            'subkey': stat['subkey'] if 'key' in stat else None,
            'data': value,
            'settings': stat,
          })
        elif 'aggregateByKey' in sd and 'value' in sd:
          xs = Tools.map(xs, '$.' + sd['selector'])
          xs = Tools.filterFn(xs, lambda x: len(x) > 0)
          rs = {}
          for x in xs:
            for x2 in x:
              key = Tools.apply(x2, '$.' + sd['aggregateByKey'])
              def _collect(selector, defaultValue):
                value = Tools.apply(x2, '$.' + selector)
                return value if value or defaultValue is None else defaultValue
              if isinstance(sd['value'], str):
                value = _collect(sd['value'], sd['defaultValue'] if 'defaultValue' in sd else None)
              else:
                value = {}
                for k, selector in sd['value'].items():
                  value[k] = _collect(selector, sd['defaultValue'][k] if 'defaultValue' in sd and k in sd['defaultValue'] else None)
              if value is not None:
                if key not in rs:
                  rs[key] = []
                rs[key].append(value)
          for key, r in rs.items():
            result.append({
              'key': stat['key'] if 'key' in stat else None,
              'subkey': (stat['subkey'] + '-' if 'subkey' in stat else '') + key,
              'subtitleAddition': key,
              'data': r,
              'settings': stat,
            })
    return result
  def statisticsData(self):
    statistics = GeneralSettings().get('statistics')
    statistics = Statistics._flattenStatistics(statistics)
    data = Statistics._readData()
    statisticsKeys = Statistics._computeStatisticsKeys(data)
    return Statistics._prepareData(data, statisticsKeys, statistics)



from Empiric.internal.StatisticsTools import StatisticsTools

class Step():
  def __init__(self, m):
    self._m = m
    self._settings = None
    self._statistics = None
  def _storeSettings(self, statistics=None, defaultStatistics=None, appendToDefaultStatistics=True, hideStatisticsByDefault=False, **kwargs):
    self._settings = kwargs
    self._statistics = StatisticsTools.mergeWithDefaultStatistics(statistics, defaultStatistics, appendToDefaultStatistics, hideStatisticsByDefault)
  def run(self, raiseError=False, **kwargs):
    self._storeSettings(**kwargs)
    return self._m.runStep(self, raiseError=raiseError) if self._m is not None else None
  def save(self, result):
    return self._m.saveCurrentStep(result)
  def settings(self):
    return self._settings
  def statistics(self):
    return self._statistics



from Empiric.Mode import MODE

import os
import random
import string

class AccessCodes():
  _accessCodeLength = 8
  _fileAccessCodes = 'access-codes.csv'
  _defaultAccessCode = '_'
  def __init__(self, mode, numberOfAccessCodes):
    self._mode = mode
    self._accessCodes = []
    if self._mode == MODE.USE_ACCESS_CODES:
      self._accessCodes = None
      if os.path.exists(AccessCodes._fileAccessCodes):
        print(' * Load list of access codes')
        with open(AccessCodes._fileAccessCodes, 'r') as f:
          self._accessCodes = f.read().splitlines()
      if self._accessCodes is None or len(self._accessCodes) < numberOfAccessCodes:
        print(' * Generate list of access codes' if self._accessCodes is None else ' * Extend list of access codes')
        if self._accessCodes is None:
          self._accessCodes = []
        while len(self._accessCodes) < numberOfAccessCodes:
          accessCode = AccessCodes.newAccessCode()
          if accessCode not in self._accessCodes:
            self._accessCodes.append(accessCode)
        with open(AccessCodes._fileAccessCodes, 'w') as f:
          f.write('\n'.join(self._accessCodes))
  def exists(self, accessCode):
    if self._mode == MODE.LOCAL:
      return accessCode == AccessCodes._defaultAccessCode
    elif self._mode == MODE.USE_ACCESS_CODES:
      return accessCode in self._accessCodes
    elif self._mode == MODE.NO_ACCESS_CODES:
      return True
  @staticmethod
  def newAccessCode():
    x = ''.join(random.choice(string.ascii_letters) for _ in range(0, AccessCodes._accessCodeLength))
    for s in ['login', 'logout', 'save', 'settings', 'statistics', 'data']:
      if x.startswith(s):
        x = AccessCodes.newAccessCode()
    return x
  @staticmethod
  def defaultAccessCode():
    return AccessCodes._defaultAccessCode



from flask import redirect, render_template, request, url_for
from flask_login import LoginManager, login_required, login_user, logout_user, UserMixin

class Authenticate:
  @staticmethod
  def init(app, statisticsPassword):
    app.config.update(SECRET_KEY='Empiric!', USE_SESSION_FOR_NEXT=True)
    loginManager = LoginManager()
    loginManager.init_app(app)
    loginManager.login_view = 'login'
    defaultUser = 'statistics'
    class User(UserMixin):
      def __init__(self, username):
        self.id = username
      def __repr__(self):
        return self.username
    @app.route('/login', methods=['GET', 'POST'])
    def login():
      if request.method == 'POST':
        if 'password' in request.form and request.form['password'] == statisticsPassword:
          login_user(User(defaultUser))
          return redirect(url_for('statistics'))
        else:
          return render_template('login.html', message='Wrong password. Please try again.')
      elif request.method == 'GET':
        return render_template('login.html')
      return render_template('login.html', message='Bad login request.')
    @app.route('/logout')
    @login_required
    def logout():
      logout_user()
      return redirect(url_for('login'))
    @loginManager.user_loader
    def loader_user(username):
      return User(username)
